@@ -4,6 +4,7 @@ import argparse
 import configparser
 import sys
 import time
+from typing import Any
 
 import requests
 
@@ -17,7 +18,7 @@ import cr_api_client.scenario_helper as scenario_helper
 #
 # 'status' related functions
 #
-def status_handler(args):
+def status_handler(args: Any) -> None:
     """Get platform status."""
 
     print("[+] Platform status")
@@ -72,7 +73,7 @@ def status_handler(args):
 #
 # 'init' related functions
 #
-def init_handler(args):
+def init_handler(args: Any) -> None:
     """Process initialization of mysql db and snapshots path."""
 
     print(
@@ -87,7 +88,7 @@ def init_handler(args):
 #
 # 'basebox_list' related functions
 #
-def basebox_list_handler(args):
+def basebox_list_handler(args: Any) -> None:
     """List available baseboxes, for use in simulations."""
     print("[+] List of available baseboxes")
     baseboxes = core_api_client.fetch_baseboxes()
@@ -115,7 +116,7 @@ def basebox_list_handler(args):
 #
 # 'simu_create' simulation related functions
 #
-def simu_create_handler(args):
+def simu_create_handler(args: Any) -> None:
     """Process YAML configuration file and request core API to create a
     new simulation.
 
@@ -127,50 +128,112 @@ def simu_create_handler(args):
     # Compute elpased time
     t1 = time.time()
 
-    id_simulation = core_helper.simu_create(architecture_file)
-    print("Created simulation ID: '{}'".format(id_simulation))
-
-    t2 = time.time()
-    time_elapsed = t2 - t1
-    print("[+] Time elapsed: {0:.2f} seconds".format(time_elapsed))
+    try:
+        id_simulation = core_helper.simu_create(architecture_file)
+        print("Created simulation ID: '{}'".format(id_simulation))
+    except Exception as e:
+        print(f"Error when creating new simulation: '{e}'")
+        sys.exit(1)
+    finally:
+        t2 = time.time()
+        time_elapsed = t2 - t1
+        print("[+] Time elapsed: {0:.2f} seconds".format(time_elapsed))
 
 
 #
-# 'aging_execute' related functions
+# 'provisioning_execute' related functions
 #
-def aging_execute_handler(args):
-    """Process YAML configuration file and execute a new aging
+def provisioning_execute_handler(args: Any) -> None:
+    """Process YAML configuration file and execute a new provisioning
     chronology (generate + play)).
 
     """
-    # Parameters
-    aging_configuration_file = args.aging_configuration_file
-    id_simulation = args.id_simulation
 
-    provisioning_helper.aging_execute(aging_configuration_file, id_simulation)
+    # Parameters
+    id_simulation = args.id_simulation
+    provisioning_configuration_file = args.provisioning_configuration_file
+    debug = args.debug_mode
+
+    try:
+        provisioning_helper.provisioning_execute(
+            id_simulation, provisioning_configuration_file, debug=debug
+        )
+    except Exception as e:
+        print(f"Error during provisioning: '{e}'")
+        sys.exit(1)
+
+
+#
+# 'provisioning_ansible' related functions
+#
+def provisioning_ansible_handler(args: Any) -> None:
+    """Apply ansible playbook on targets."""
+
+    # Parameters
+    id_simulation = args.id_simulation
+    provisioning_playbook_path = args.provisioning_playbook_path
+    provisioning_target_role = args.provisioning_target_role
+    provisioning_target_system_type = args.provisioning_target_system_type
+    provisioning_target_operating_system = args.provisioning_target_operating_system
+    provisioning_target_name = args.provisioning_target_name
+    debug = args.debug_mode
+
+    # TODO: check that only one target type is defined
+
+    # TODO: support other target types
+    if provisioning_target_operating_system is not None:
+        raise NotImplementedError("target_operating_system not yet supported")
+    if provisioning_target_system_type is not None:
+        raise NotImplementedError("target_system_type not yet supported")
+    if provisioning_target_role is not None:
+        raise NotImplementedError("target_role not yet supported")
+
+    try:
+        provisioning_helper.provisioning_ansible(
+            id_simulation,
+            provisioning_playbook_path,
+            target_role=provisioning_target_role,
+            target_system_type=provisioning_target_system_type,
+            target_operating_system=provisioning_target_operating_system,
+            target_name=provisioning_target_name,
+            debug=debug,
+        )
+    except Exception as e:
+        print(f"Error during provisioning: '{e}'")
+        sys.exit(1)
 
 
 #
 # 'scenario_play' simulation
 #
-def scenario_play_handler(args):
+def scenario_play_handler(args: Any) -> None:
     """Play scenario on targeted simulation."""
     # Parameters
     id_simulation = args.id_simulation
     scenario_path = args.scenario_path
+    file_results = args.scenario_file_results
+    debug_mode = args.scenario_debug_mode
+    speed = args.scenario_speed
 
     print(
         "[+] Playing scenario '{}' on simulation id '{}'".format(
             scenario_path, id_simulation
         )
     )
-    scenario_helper.scenario_play(id_simulation, scenario_path)
+
+    try:
+        scenario_helper.scenario_play(
+            id_simulation, scenario_path, debug_mode, speed, file_results
+        )
+    except Exception as e:
+        print(f"Error when playing scenario: '{e}'")
+        sys.exit(1)
 
 
 #
 # 'scenario_status' simulation
 #
-def scenario_status_handler(args):
+def scenario_status_handler(args: Any) -> None:
     """Get scenario status on targeted simulation."""
     # Parameters
     id_simulation = args.id_simulation
@@ -183,27 +246,31 @@ def scenario_status_handler(args):
 #
 # 'simu_run' simulation
 #
-def simu_run_handler(args):
+def simu_run_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
-    use_vm_time = not args.use_current_time
+    use_install_time = args.use_install_time
 
     # Compute elpased time
     t1 = time.time()
 
-    core_helper.simu_run(id_simulation, use_vm_time)
-
-    t2 = time.time()
-    time_elapsed = t2 - t1
-    print("[+] Time elapsed: {0:.2f} seconds".format(time_elapsed))
-
-    print("[+] Simulation is running...")
+    try:
+        core_helper.simu_run(id_simulation, use_install_time)
+    except Exception as e:
+        print(f"Error when starting simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("[+] Simulation is running...")
+    finally:
+        t2 = time.time()
+        time_elapsed = t2 - t1
+        print("[+] Time elapsed: {0:.2f} seconds".format(time_elapsed))
 
 
 #
 # 'simu_status' of simulation
 #
-def simu_status_handler(args):
+def simu_status_handler(args: Any) -> None:
     # Parameters
     requested_simulation_id = args.id_simulation
 
@@ -274,121 +341,183 @@ def simu_status_handler(args):
 #
 # 'simu_pause' simulation
 #
-def simu_pause_handler(args):
+def simu_pause_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    core_helper.simu_pause(id_simulation)
-    print("Simulation paused")
+    try:
+        core_helper.simu_pause(id_simulation)
+    except Exception as e:
+        print(f"Error when pausing simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Simulation paused")
 
 
 #
 # 'simu_unpause' simulation
 #
-def simu_unpause_handler(args):
+def simu_unpause_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    core_helper.simu_unpause(id_simulation)
-    print("Simulation unpaused")
+    try:
+        core_helper.simu_unpause(id_simulation)
+    except Exception as e:
+        print(f"Error when unpausing simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Simulation unpaused")
 
 
 #
 # 'simu_halt' simulation
 #
-def simu_halt_handler(args):
+def simu_halt_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    core_helper.simu_halt(id_simulation)
-    print("Simulation halted")
+    try:
+        core_helper.simu_halt(id_simulation)
+    except Exception as e:
+        print(f"Error when halting simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Simulation halted")
 
 
 #
 # 'simu_destroy' simulation
 #
-def simu_destroy_handler(args):
+def simu_destroy_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    core_helper.simu_destroy(id_simulation)
-    print("Simulation destroyed")
+    try:
+        core_helper.simu_destroy(id_simulation)
+    except Exception as e:
+        print(f"Error when destroying simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Simulation destroyed")
 
 
 #
 # 'simu_clone' simulation
 #
-def simu_clone_handler(args):
+def simu_clone_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    id_new_simulation = core_helper.simu_clone(id_simulation)
-    print("Simulation cloned")
-    print("Created simulation ID: '{}'".format(id_new_simulation))
+    try:
+        id_new_simulation = core_helper.simu_clone(id_simulation)
+    except Exception as e:
+        print(f"Error when cloning simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Simulation cloned")
+        print("Created simulation ID: '{}'".format(id_new_simulation))
 
 
 #
 # 'simu_tap' simulation
 #
-def simu_tap_handler(args):
+def simu_tap_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
     iface = args.iface
 
-    core_helper.simu_tap(id_simulation, iface)
-    print("Redirect network traffic to the tap interface")
+    try:
+        core_helper.simu_tap(id_simulation, iface)
+    except Exception as e:
+        print(f"Error when setting tap on simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Redirect network traffic to the tap interface")
 
 
 #
 # 'simu_untap' simulation
 #
-def simu_untap_handler(args):
+def simu_untap_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
     iface = args.iface
 
-    core_helper.simu_untap(id_simulation, iface)
-    print("Stop redirection of network traffic to the tap interface")
+    try:
+        core_helper.simu_untap(id_simulation, iface)
+    except Exception as e:
+        print(f"Error when unsetting tap on simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("Stop redirection of network traffic to the tap interface")
 
 
 #
 # 'simu_delete' simulation
 #
-def simu_delete_handler(args):
+def simu_delete_handler(args: Any) -> None:
     # Parameters
     id_simulation = args.id_simulation
 
-    core_helper.simu_delete(id_simulation)
-    print("[+] VMs destroyed")
-    print("[+] VMs snapshots deleted")
-    print("[+] Simulation deleted from database")
+    try:
+        core_helper.simu_delete(id_simulation)
+    except Exception as e:
+        print(f"Error when deleting simulation: '{e}'")
+        sys.exit(1)
+    else:
+        print("[+] VMs destroyed")
+        print("[+] VMs snapshots deleted")
+        print("[+] Simulation deleted from database")
 
 
-def set_core_api_url(core_api_url):
+#
+# 'simu_snap' simulation
+#
+def simu_snap_handler(args: Any) -> None:
+    # Parameters
+    id_simulation = args.id_simulation
+    # core generates an architecture file (yaml)
+    # that will be located to /cyber_range_stuff...
+    # This outfile is merely a copy of the generated
+    # file
+    output_file = args.output
+
+    try:
+        yaml = core_helper.simu_snap(id_simulation)
+        with open(output_file, "w") as w:
+            w.write(yaml)
+        print(f"[+] Snapshot done. Architecture file stored at {output_file}")
+    except Exception as e:
+        print(f"Error when creating snapshot for simulation: '{e}'")
+        sys.exit(1)
+
+
+def set_core_api_url(core_api_url: str) -> str:
     print("  [+] Using core API URL: {}".format(core_api_url))
     core_api_client.CORE_API_URL = core_api_url
     return core_api_url
 
 
-def set_scenario_api_url(scenario_api_url):
+def set_scenario_api_url(scenario_api_url: str) -> str:
     print("  [+] Using scenario API URL: {}".format(scenario_api_url))
     scenario_helper.SCENARIO_API_URL = scenario_api_url
     return scenario_api_url
 
 
-def set_provisioning_api_url(provisioning_api_url):
+def set_provisioning_api_url(provisioning_api_url: str) -> str:
     print("  [+] Using provisioning API URL: {}".format(provisioning_api_url))
     provisioning_helper.PROVISIONING_API_URL = provisioning_api_url
     return provisioning_api_url
 
 
-def set_redteam_api_url(redteam_api_url):
+def set_redteam_api_url(redteam_api_url: str) -> str:
     print("  [+] Using redteam API URL: {}".format(redteam_api_url))
     redteam_helper.REDTEAM_API_URL = redteam_api_url
     return redteam_api_url
 
 
-def set_cacert(cacert):
+def set_cacert(cacert: str) -> str:
     print("  [+] Using CA certs path: {}".format(cacert))
     core_api_client.CA_CERT_PATH = cacert
     scenario_helper.CA_CERT_PATH = cacert
@@ -397,7 +526,7 @@ def set_cacert(cacert):
     return cacert
 
 
-def set_cert(cert):
+def set_cert(cert: str) -> str:
     print("  [+] Using client cert path: {}".format(cert))
     core_api_client.CLIENT_CERT_PATH = cert
     scenario_helper.CLIENT_CERT_PATH = cert
@@ -406,7 +535,7 @@ def set_cert(cert):
     return cert
 
 
-def set_key(key):
+def set_key(key: str) -> str:
     print("  [+] Using client key path: {}".format(key))
     core_api_client.CLIENT_KEY_PATH = key
     scenario_helper.CLIENT_KEY_PATH = key
@@ -415,7 +544,7 @@ def set_key(key):
     return key
 
 
-def main():
+def main() -> None:
 
     print("[+] Config")
     parser = argparse.ArgumentParser()
@@ -521,10 +650,10 @@ def main():
     parser_simu_run.set_defaults(func=simu_run_handler)
     parser_simu_run.add_argument("id_simulation", type=int, help="The simulation id")
     parser_simu_run.add_argument(
-        "--use_current_time",
+        "--use_install_time",
         action="store_true",
-        dest="use_current_time",
-        help="Indicates that hypervisor time will be used to set VMs boot time",
+        dest="use_install_time",
+        help="Indicates that VM installation time will be used to set VMs boot time",
     )
 
     # 'simu_status' simulation command
@@ -572,6 +701,21 @@ def main():
         "id_simulation", type=int, help="The simulation id"
     )
 
+    # 'simu_snap' simulation command
+    parser_simu_snap = subparsers.add_parser(
+        "simu_snap", help="Get status of a simulation or all simulations"
+    )
+    parser_simu_snap.set_defaults(func=simu_snap_handler)
+    parser_simu_snap.add_argument("id_simulation", type=int, help="The simulation id")
+    parser_simu_snap.add_argument(
+        "-o",
+        "--output",
+        type=str,
+        required=True,
+        default=None,
+        help="Path to the output yaml file",
+    )
+
     # 'simu_clone' simulation command
     parser_simu_clone = subparsers.add_parser("simu_clone", help="Clone a simulation")
     parser_simu_clone.set_defaults(func=simu_clone_handler)
@@ -604,23 +748,63 @@ def main():
     parser_simu_untap.add_argument("iface", type=str, help="The tap network interface")
 
     # -----------------------
-    # --- Aging options
+    # --- Provisioning options
     # -----------------------
 
-    # 'aging_execute' command
-    parser_aging_execute = subparsers.add_parser(
-        "aging_execute", help="Execute aging chronology for a simulation"
+    # 'provisioning_execute' command
+    parser_provisioning_execute = subparsers.add_parser(
+        "provisioning_execute", help="Execute provisioning chronology for a simulation"
     )
-    parser_aging_execute.set_defaults(func=aging_execute_handler)
-    parser_aging_execute.add_argument(
+    parser_provisioning_execute.set_defaults(func=provisioning_execute_handler)
+    parser_provisioning_execute.add_argument(
         "id_simulation", type=int, help="The simulation id"
     )
-    parser_aging_execute.add_argument(
+    parser_provisioning_execute.add_argument(
         "-c",
         action="store",
         required=True,
-        dest="aging_configuration_file",
-        help="Input path of aging configuration",
+        dest="provisioning_configuration_file",
+        help="Input path of provisioning configuration",
+    )
+
+    # 'provisioning_ansible' command
+    parser_provisioning_ansible = subparsers.add_parser(
+        "provisioning_ansible", help="Apply ansible playbook(s) on targets"
+    )
+    parser_provisioning_ansible.set_defaults(func=provisioning_ansible_handler)
+    parser_provisioning_ansible.add_argument(
+        "id_simulation", type=int, help="The simulation id"
+    )
+    parser_provisioning_ansible.add_argument(
+        "-c",
+        action="store",
+        required=True,
+        dest="provisioning_playbook_path",
+        help="Input directory containing ansible playbook(s)",
+    )
+    parser_provisioning_ansible.add_argument(
+        "-r",
+        action="store",
+        dest="provisioning_target_role",
+        help="Role used to filter targets ('client', 'activate_directory', 'file_server', 'admin', ...)",
+    )
+    parser_provisioning_ansible.add_argument(
+        "-s",
+        action="store",
+        dest="provisioning_target_system_type",
+        help="System type used to filter targets ('linux', 'windows')",
+    )
+    parser_provisioning_ansible.add_argument(
+        "-o",
+        action="store",
+        dest="provisioning_target_operating_system",
+        help="Operating system used to filter targets ('Windows 7', 'Windows 10', 'Debian', 'Ubuntu', ...)",
+    )
+    parser_provisioning_ansible.add_argument(
+        "-n",
+        action="store",
+        dest="provisioning_target_name",
+        help="Machine name used to filter targets",
     )
 
     # -----------------------
@@ -639,11 +823,36 @@ def main():
         "-i",
         action="store",
         nargs="?",
-        # required=True,
+        required=True,
         dest="scenario_path",
         help="Path of the scenario to play",
     )
 
+    parser_scenario_play.add_argument(
+        "-o",
+        action="store",
+        required=False,
+        dest="scenario_file_results",
+        help="Absolute name of scenario results (json format)",
+    )
+
+    parser_scenario_play.add_argument(
+        "-d",
+        action="store",
+        required=False,
+        dest="scenario_debug_mode",
+        default="off",
+        help="Debug mode ('off', 'on', 'full')",
+    )
+    parser_scenario_play.add_argument(
+        "-t",
+        action="store",
+        required=False,
+        dest="scenario_speed",
+        default="normal",
+        help="scenario speed ('slow', 'normal', 'fast')",
+    )
+
     # 'scenario_status' command
     parser_scenario_status = subparsers.add_parser(
         "scenario_status", help="Get scenario status on a simulation"
@@ -662,7 +871,8 @@ def main():
         for k, v in config.items("DEFAULT"):
             parser.parse_args([str(k), str(v)], args)
 
-    # Parse remaining args from command line (overriding potential config file parameters)
+    # Parse remaining args from command line (overriding potential config file
+    # parameters)
     args = parser.parse_args(left_argv, args)
 
     args.func(args)
